
#include "file_write.h"

extern struct RECV_BUF recv_buf;
extern int Exit;
extern int Save;
extern FILE *fp_log;


//==========================================================================
// Data Receive Worker Thread
//==========================================================================
DWORD WINAPI FileWriteWorker(LPVOID arg) {
	//-----------------------------------------------------------------
	// Variables
	//-----------------------------------------------------------------
	int byte_count;
	unsigned char *ptr;
	int bank;
	int frame_num;
	char filename[200];
	int file_index;
	HEADER_INFO header_info;
	int init;
	unsigned char pix_buf[8];

	cv::Mat matImgA;
	cv::Mat matImgB;
	cv::Mat *matImgActive = &matImgA;

	cv::namedWindow("test", CV_WINDOW_AUTOSIZE);

	//-----------------------------------------------------------------
	// Initialize
	//-----------------------------------------------------------------
	byte_count = 0;
	bank = 0;
	frame_num = 0;
	file_index = 0;
	init = 0;
	header_info.found = 0;
	header_info.state = 0;

	//-----------------------------------------------------------------
	// Data Reception Loop
	//-----------------------------------------------------------------
	int row, col;
	while (1) {
		// wait for data to be ready in the current bank
		while (recv_buf.flag[bank] != 1) {
			//printf("[%d]", recv_buf.flag[bank]);
			Sleep(10);
		}

		// pointer for data head
		ptr = (unsigned char*)recv_buf.buf[bank];

		// data read loop
		for (int index = 0; index < recv_buf.len[bank]; index++) {

			if (header_info.found == 0) {

				// search for header
				SearchHeader(ptr, &header_info);

				// header is found
				if (header_info.found == 1) {
					if (init == 0) {
						// this process will run only once
						printf(
							"tyep1 %d, type2 %d, total(h:%d, w:%d), active(h:%d, w:%d)\n",
							header_info.type1, header_info.type2,
							header_info.tot_height, header_info.tot_width,
							header_info.act_height, header_info.act_width
						);
						fprintf(
							fp_log,
							"tyep1 %d, type2 %d, total(h:%d, w:%d), active(h:%d, w:%d)\n",
							header_info.type1, header_info.type2,
							header_info.tot_height, header_info.tot_width,
							header_info.act_height, header_info.act_width
						);
						matImgA = cv::Mat(header_info.act_height, header_info.act_width, CV_8UC3);
						matImgB = cv::Mat(header_info.act_height, header_info.act_width, CV_8UC3);
						init = 1;
					}
					row = 0;
					col = 0;
					fprintf(fp_log, "header found\n");
				}
				ptr++;
			}
			else {
				// acquire image data
				// store one byte at a time until pixel boundary
				pix_buf[byte_count++] = *ptr++;

				if (byte_count == header_info.bytes_per_chunk) {
					// pixel boundary reached
					byte_count = 0;

					// convert to pixel value according to the frame type
					// then store to cv::Mat structure
					if ((row < header_info.act_height) && (col < header_info.act_width)) {
						switch (header_info.type1) {
						case 0:
							// Frame Grabber
							switch (header_info.type2) {
							case 0:
								// L-R Interleaved
								GetPixelBgr(matImgActive, row, col, pix_buf, header_info.act_width);
								break;
							default:
								break;
							}
							break;
						case 1:
							// Block Matching
							GetDispInteger(matImgActive, row, col, pix_buf);
							break;
						case 2:
							// Stereo Rectification
							GetLum(matImgActive, row, col, pix_buf);
							break;
						case 3:
							// X-Sobel
							GetLumLR(matImgActive, row, col, pix_buf, header_info.act_width);
							break;
						default:
							break;
						}
					}

					col += header_info.pixels_per_chunk;

					// end of column
					if (col == header_info.tot_width) {
						col = 0;
						row++;

						// end of row
						if (row == header_info.tot_height) {
							row = 0;

							// frame received
							printf("*");

							// store image
							if ((frame_num == 0) | (Save == 1)) {
								sprintf(filename, "img_%03d.bmp", file_index++);
								cv::imwrite(filename, *matImgActive);
								Save = 0;
								fprintf(fp_log, "image saved\n");
							}

							// display image
							cv::imshow("test", *matImgActive);
							cv::waitKey(1);

							// clear image buffer (for debug)
							for (int row = 0; row < matImgActive->rows; row++) {
								for (int col = 0; col < matImgActive->cols; col++) {
									matImgActive->at<cv::Vec3b>(row, col) = cv::Vec3b(0, 0, 0);
								}
							}

							// go to next frame
							frame_num++;
							header_info.found = 0;

							// switch to another Mat structure to display
							matImgActive = ((frame_num % 2) == 0) ? &matImgA : &matImgB;

							fprintf(
								fp_log,
								"frame_num changed to %d (%d %d %d %d %d %d %d %d) len %d index %d\n",
								frame_num,
								pix_buf[0], pix_buf[1], pix_buf[2], pix_buf[3], pix_buf[4], pix_buf[5], pix_buf[6], pix_buf[7],
								recv_buf.len[bank], index
							);
						} // end of row
					} // end of column
				} // pixel boundary
			} // image data acquisiton
		} // // data read loop

		// buffer is empty, clear the flag
		recv_buf.flag[bank] = 0;

		// go to next bank
		if (bank == NUM_BUF - 1) {
			bank = 0;
		}
		else {
			bank++;
		}

		// leave if exit flag is set
		if (Exit == 1) {
			printf(         "\nFileWriteWorker Exit\n");
			fprintf(fp_log, "\nFileWriteWorker Exit\n");
			break;
		}
	} // while loop

	return 0;
}

//==========================================================================
// SearchHeader
//--------------------------------------------------------------------------
// 0: Preamble (FFFF_0000)
// 1: Identifier (4E75_4754, "NuGT")
// 2: total length in bytes
// 3: data type (type1[15:0], type2[15:0])
// 4: auxiliary 0 (total height[15:0], width[15:0])
// 5: auxiliary 1 (active height[15:0], width[15:0])
// 6: auxiliary 2 (reserved)
// 7: auxiliary 3 (reserved)
//--------------------------------------------------------------------------
// type1	type2
//     0	    0 : Frame Grabber (L-R Interleaved)
//     0     else : Frame Grabber (reserved)
//     1      n.a : Block Mathcing (depth map)
//     2     else : Stereo Rectification
//--------------------------------------------------------------------------
// aux0-3 are dependant on the data type.
//==========================================================================
void SearchHeader(
	unsigned char *ptr,
	struct HEADER_INFO *header_info
) {
	// beware of endianness
	switch (header_info->state) {
		//-----------------------------------------------------------------
		// Preamble (0xFFFF0000)
		//-----------------------------------------------------------------
		case 0:
			header_info->state = (*ptr == 0x00) ? header_info->state + 1 : 0;
			break;
		case 1:
			header_info->state = (*ptr == 0x00) ? header_info->state + 1 : 0;
			break;
		case 2:
			header_info->state = (*ptr == 0xFF) ? header_info->state + 1 : 0;
			break;
		case 3:
			header_info->state = (*ptr == 0xFF) ? header_info->state + 1 : 0;
			break;
		//-----------------------------------------------------------------
		// Identifier ('NuGT' = 0x4E754754)
		//-----------------------------------------------------------------
		case 4:
			header_info->state = (*ptr == 'T') ? header_info->state + 1 : 0;
			break;
		case 5:
			header_info->state = (*ptr == 'G') ? header_info->state + 1 : 0;
			break;
		case 6:
			header_info->state = (*ptr == 'u') ? header_info->state + 1 : 0;
			break;
		case 7:
			header_info->state = (*ptr == 'N') ? header_info->state + 1 : 0;
			break;
		//-----------------------------------------------------------------
		// Total length in bytes
		//-----------------------------------------------------------------
		case 8:
			header_info->tot_len = *ptr;
			header_info->state++;
			break;
		case 9:
			header_info->tot_len += (*ptr << 8);
			header_info->state++;
			break;
		case 10:
			header_info->tot_len += (*ptr << 16);
			header_info->state++;
			break;
		case 11:
			header_info->tot_len += (*ptr << 24);
			header_info->state++;
			break;
		//-----------------------------------------------------------------
		// Data type
		//-----------------------------------------------------------------
		case 12:
			header_info->type2 = *ptr;
			header_info->state++;
			break;
		case 13:
			header_info->type2 += (*ptr << 8);
			header_info->state++;
			break;
		case 14:
			header_info->type1 = *ptr;
			header_info->state++;
			break;
		case 15:
			header_info->type1 += (*ptr << 8);
			header_info->state++;
			break;
		//-----------------------------------------------------------------
		// Auxiliary 0
		//-----------------------------------------------------------------
		case 16:
			header_info->aux0 = *ptr;
			header_info->state++;
			break;
		case 17:
			header_info->aux0 += (*ptr << 8);
			header_info->state++;
			break;
		case 18:
			header_info->aux0 += (*ptr << 16);
			header_info->state++;
			break;
		case 19:
			header_info->aux0 += (*ptr << 24);
			header_info->state++;
			break;
		//-----------------------------------------------------------------
		// Auxiliary 1
		//-----------------------------------------------------------------
		case 20:
			header_info->aux1 = *ptr;
			header_info->state++;
			break;
		case 21:
			header_info->aux1 += (*ptr << 8);
			header_info->state++;
			break;
		case 22:
			header_info->aux1 += (*ptr << 16);
			header_info->state++;
			break;
		case 23:
			header_info->aux1 += (*ptr << 24);
			header_info->state++;
			break;
		//-----------------------------------------------------------------
		// Auxiliary 2
		//-----------------------------------------------------------------
		case 24:
			header_info->aux2 = *ptr;
			header_info->state++;
			break;
		case 25:
			header_info->aux2 += (*ptr << 8);
			header_info->state++;
			break;
		case 26:
			header_info->aux2 += (*ptr << 16);
			header_info->state++;
			break;
		case 27:
			header_info->aux2 += (*ptr << 24);
			header_info->state++;
			break;
		//-----------------------------------------------------------------
		// Auxiliary 3
		//-----------------------------------------------------------------
		case 28:
			header_info->aux3 = *ptr;
			header_info->state++;
			break;
		case 29:
			header_info->aux3 += (*ptr << 8);
			header_info->state++;
			break;
		case 30:
			header_info->aux3 += (*ptr << 16);
			header_info->state++;
			break;
		case 31:
			header_info->aux3 += (*ptr << 24);
			header_info->state++;
			break;
		//-----------------------------------------------------------------
		// end of header
		//-----------------------------------------------------------------
		case 63:
			header_info->state = 0;
			header_info->found = 1;

			// settings according to frame type
			switch (header_info->type1) {
			case 0: // Frame Grabber

				// image size
				header_info->tot_height = (header_info->aux0 & 0xFFFF0000) >> 16;
				header_info->tot_width  =  header_info->aux0 & 0x0000FFFF;
				header_info->act_height = (header_info->aux1 & 0xFFFF0000) >> 16;
				header_info->act_width  =  header_info->aux1 & 0x0000FFFF;

				switch (header_info->type2) {
				case 0:
					// L-R Interleaved
					// --- L --- --- R ---
					// Y Cb Y Cr Y Cb Y Cr
					header_info->bytes_per_chunk = 8;
					header_info->pixels_per_chunk = 4;
					break;
				default:
					printf("undefined type2 %d\n", header_info->type2);
					break;
				}
				break;
			case 1: // Block Matching

				// image size
				header_info->tot_height = (header_info->aux0 & 0xFFFF0000) >> 16;
				header_info->tot_width  = header_info->aux0 & 0x0000FFFF;
				header_info->act_height = (header_info->aux1 & 0xFFFF0000) >> 16;
				header_info->act_width  = header_info->aux1 & 0x0000FFFF;

				// 4 bytes per pixel
				header_info->bytes_per_chunk = 4;
				header_info->pixels_per_chunk = 1;
				break;
			case 2: // Stereo Rectification

				// image size
				header_info->tot_height = (header_info->aux0 & 0xFFFF0000) >> 16;
				header_info->tot_width = header_info->aux0 & 0x0000FFFF;
				header_info->act_height = (header_info->aux1 & 0xFFFF0000) >> 16;
				header_info->act_width = header_info->aux1 & 0x0000FFFF;

				// 1 byte per pixel
				// process 4 bytes at once for endian conversion
				header_info->bytes_per_chunk = 4;
				header_info->pixels_per_chunk = 4;
				break;
			case 3: // X-Sobel

				// image size
				header_info->tot_height = (header_info->aux0 & 0xFFFF0000) >> 16;
				header_info->tot_width = header_info->aux0 & 0x0000FFFF;
				header_info->act_height = (header_info->aux1 & 0xFFFF0000) >> 16;
				header_info->act_width = header_info->aux1 & 0x0000FFFF;

				// lum only, L-R interleaved
				header_info->bytes_per_chunk = 4;
				header_info->pixels_per_chunk = 4;
				break;
			default:
				printf("undefined type1 %d\n", header_info->type1);
				break;
			}
			
			break;
		default:
			header_info->state++;
			break;
	}
}

void GetLum(cv::Mat *matImgActive, int row, int col, unsigned char *pix_buf) {
	cv::Vec3b bgr;

	// convert lum to BGR
	// big endian to little endian
	bgr = cv::Vec3b(pix_buf[3], pix_buf[3], pix_buf[3]);
	matImgActive->at<cv::Vec3b>(row, col) = bgr;

	bgr = cv::Vec3b(pix_buf[2], pix_buf[2], pix_buf[2]);
	matImgActive->at<cv::Vec3b>(row, col + 1) = bgr;

	bgr = cv::Vec3b(pix_buf[1], pix_buf[1], pix_buf[1]);
	matImgActive->at<cv::Vec3b>(row, col + 2) = bgr;

	bgr = cv::Vec3b(pix_buf[0], pix_buf[0], pix_buf[0]);
	matImgActive->at<cv::Vec3b>(row, col + 3) = bgr;

	return;
}

void GetLumLR(cv::Mat *matImgActive, int row, int col, unsigned char *pix_buf, int W) {
	cv::Vec3b bgr;

	int col_l = col / 2; // left half
	int col_r = col_l + W / 2; // right half

	// convert lum to BGR
	// big endian to little endian
	bgr = cv::Vec3b(pix_buf[3], pix_buf[3], pix_buf[3]);
	matImgActive->at<cv::Vec3b>(row, col_l) = bgr;

	bgr = cv::Vec3b(pix_buf[2], pix_buf[2], pix_buf[2]);
	matImgActive->at<cv::Vec3b>(row, col_r) = bgr;

	bgr = cv::Vec3b(pix_buf[1], pix_buf[1], pix_buf[1]);
	matImgActive->at<cv::Vec3b>(row, col_l + 1) = bgr;

	bgr = cv::Vec3b(pix_buf[0], pix_buf[0], pix_buf[0]);
	matImgActive->at<cv::Vec3b>(row, col_r + 1) = bgr;

	return;
}

void GetDispInteger(cv::Mat *matImgActive, int row, int col, unsigned char *pix_buf) {
	// most significant byte contains integer part of disparity
	// convert this byte to BGR, store to image buffer
	cv::Vec3b bgr = cv::Vec3b(pix_buf[3], pix_buf[3], pix_buf[3]);
	matImgActive->at<cv::Vec3b>(row, col) = bgr;

	return;
}

void GetPixelBgr(cv::Mat *matImgActive, int row, int col, unsigned char *pix_buf, int W) {

	cv::Vec3b bgr;

	int col_l = col / 2; // left half
	int col_r = col_l + W / 2; // right half

	// 1st Camera / 1st pixel
	YUVtoBGR(pix_buf[0], pix_buf[2], pix_buf[6], &bgr);
	matImgActive->at<cv::Vec3b>(row, col_l) = bgr;

	// 1st Camera / 2nd pixel
	YUVtoBGR(pix_buf[4], pix_buf[2], pix_buf[6], &bgr);
	matImgActive->at<cv::Vec3b>(row, col_l + 1) = bgr;

	// 2nd Camera / 1st pixel
	YUVtoBGR(pix_buf[1], pix_buf[3], pix_buf[7], &bgr);
	matImgActive->at<cv::Vec3b>(row, col_r) = bgr;

	// 2nd Camera / 2nd pixel
	YUVtoBGR(pix_buf[5], pix_buf[3], pix_buf[7], &bgr);
	matImgActive->at<cv::Vec3b>(row, col_r + 1) = bgr;

	return;
}

void YUVtoBGR(unsigned char y, unsigned char u, unsigned char v, cv::Vec3b *bgr) {
	double b, g, r;

	b = y + (u - 128);
	g = y - (u - 128) * 0.186 - (v - 128) * 0.508;
	r = y + (v - 128);

	if (b > 255) b = 255;
	else if (b < 0) b = 0;
	if (g > 255) g = 255;
	else if (g < 0) g = 0;
	if (r > 255) r = 255;
	else if (r < 0) r = 0;

	bgr->val[0] = (unsigned char)b;
	bgr->val[1] = (unsigned char)g;
	bgr->val[2] = (unsigned char)r;

	return;
}
