
#include "tcp_recv.h"

int BufferReady;
struct RECV_BUF recv_buf;

extern int Exit;
extern FILE *fp_log;
extern int Port;


//***************************************************************************
// PrintErrorMessage
//---------------------------------------------------------------------------
// run WSAGetLastError to acquire error code, display corresponding error
// message.
//***************************************************************************
DWORD PrintErrorMessage() {
	// acquire error code
	DWORD errorcode = WSAGetLastError();

	// acquire error message string
	LPVOID lpMsgBuf;
	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER  // function allocates a buffer
		| FORMAT_MESSAGE_FROM_SYSTEM    // search the system message-table
		| FORMAT_MESSAGE_IGNORE_INSERTS,// Insert sequences in the message definition are to be ignored
		NULL,
		errorcode,
		1033, // US English
		(LPTSTR)&lpMsgBuf, // A pointer to a buffer that receives the null-terminated string
		0,
		NULL
	);

	// display
	printf(         "%d:%s\n", errorcode, (char*)lpMsgBuf);
	fprintf(fp_log, "%d:%s\n", errorcode, (char*)lpMsgBuf);

	LocalFree(lpMsgBuf);

	return errorcode;
}

//***************************************************************************
// TcpRecvWorker
//---------------------------------------------------------------------------
// this worker thread handles TCP communication
//***************************************************************************
DWORD WINAPI TcpRecvWorker(LPVOID arg) {
	WSADATA wsaData;
	struct sockaddr_in local_addr;
	struct sockaddr_in client;
	SOCKET sock, sock2;
	clock_t st_time = 0;
	clock_t el_time;
	double data_rate;
	BOOL yes = 1;
	//int tot_len = 0;
	long long tot_len = 0;
	int len;
	int frame = 0;
	int bank = 0;
	int ExitSelf = 0;

	//=======================================================================
	// TCP connection link
	//=======================================================================
	// initialize winsock2
	printf("initiates use of the Winsock DLL\n");
	if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0) {
		PrintErrorMessage();
		return -1;
	}

	// create socket
	printf("creates a socket\n");
	//sock = socket(AF_INET, SOCK_STREAM, 0);
	sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock == INVALID_SOCKET) {
		PrintErrorMessage();
		return -1;
	}

	// necessary to reuse once closed IP address
	printf("allows reuse address\n");
	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (const char *)&yes, sizeof(yes)) != 0) {
		PrintErrorMessage();
		return -1;
	}

	// local terminals from which accept connection
	local_addr.sin_family = AF_INET;
	local_addr.sin_port = htons(Port);
	local_addr.sin_addr.S_un.S_addr = INADDR_ANY;

	printf("associates a local address with a socket\n");
	if (bind(sock, (struct sockaddr *)&local_addr, sizeof(local_addr)) != 0) {
		PrintErrorMessage();
		return -1;
	}

	// places a socket in listening state
	printf("places a socket in listening state\n");
	if (listen(sock, 5) != 0) {
		PrintErrorMessage();
		return -1;
	}

	// waiting for an incoming connection attempt on a socket
	printf("waiting for an incoming connection attempt on a socket\n");
	len = sizeof(client);
	sock2 = accept(sock, (struct sockaddr *)&client, &len);
	if (sock2 == -1) {
		PrintErrorMessage();
		return -1;
	}
	else {
		printf("accepted connection from %s:%d\n", inet_ntoa(client.sin_addr), ntohs(client.sin_port));
	}

	//=======================================================================
	// data reception main loop
	//=======================================================================
	while (1) {
		// data remains in newly assinged bank
		// -> overrun error
		if (recv_buf.flag[bank] == 1) {
			printf(         "overrun at bank %d\n", bank);
			fprintf(fp_log, "overrun at bank %d\n", bank);
		}

		// data reception
		recv_buf.len[bank] = recv(sock2, recv_buf.buf[bank], BUFLEN, 0);
		if (recv_buf.len[bank] < 0) {
			printf(         "recv returned negative value\n");
			fprintf(fp_log, "recv returned negative value\n");
			DWORD errorcode = PrintErrorMessage();
			if (errorcode == 10054) {
				break;
			}
		}
		else if (recv_buf.len[bank] == 0) {
			printf("connection closed\n");
			break;
		}
		else {
			// count elapsed time
			if (tot_len == 0) {
				// do once, remember the start time
				printf("recv start\n");
				//fprintf(fp_log, "recv start\n");
				st_time = clock();
			}
			else {
				el_time = clock();
				el_time -= st_time;
			}

			// record the number of total received bytes
			unsigned long *tmpl = (unsigned long*)recv_buf.buf[bank];
			tot_len += recv_buf.len[bank];
			//printf(         "%6d bytes received by %2d (total %6d): %08lx\n", recv_buf.len[bank], bank, tot_len, *tmpl);
			//fprintf(fp_log, "%6d bytes received by %2d (total %6d): %08lx\n", recv_buf.len[bank], bank, tot_len, *tmpl);

			// this bank now contains data
			recv_buf.flag[bank] = 1;

			// go to next bank
			if (bank == NUM_BUF - 1) {
				bank = 0;
			}
			else {
				bank++;
			}

			// leave the process when the specified number of frames are received.
			// continue when NUM_FRAME=0.
			if ((NUM_FRAME != 0) && (tot_len >= FRAME_SIZE * NUM_FRAME)) {
				ExitSelf = 1;
				break;
			}
		}

		if ((Exit == 1) || (ExitSelf == 1)) {
			printf("\nTcpRecvWorker Exit\n");
			break;
		}
	}

	printf(         "total bytes received %lld\n", tot_len);
	fprintf(fp_log, "total bytes received %lld\n", tot_len);
	data_rate = (double)(tot_len / (double)(el_time * 1000));
	printf(         "%.2f MB/s\n", data_rate);
	fprintf(fp_log, "%.2f MB/s\n", data_rate);

	closesocket(sock2);

	// clean up winsock2
	WSACleanup();

	return 0;
}
